/***************************************************************************/
#include <stdio.h>
#include <float.h>
#include <assert.h>
#include "FibonacciHeap.h"
#include "FibonacciHeapGlobal.h"
#include "xmenus.h"
#include "const.h"
/***************************************************************************/
FibonacciHeap::FibonacciHeap(void) 
{
	min=NULL;
	max_rank=0;
	nroot=0;
	nnode=0;

#ifdef PREALLOCATE
	if (nchunk>=NCHUNKS) {
		put_error("Error:Too many simultaneous heaps");
		return;
	}
	myChunk=nchunk;
	chunk[myChunk].nglobal_fnode=0;	// So far we have used 0 nodes
	nchunk++;
#endif
}
/******************************************************************************/
// Deletes the whole structures
FibonacciHeap::~FibonacciHeap(void)
{
#ifdef PREALLOCATE
	nchunk--;	// Since we dont need it
#else
	// Old way
	FNODE *root,*next;

	// Destroy all root nodes
	next=min;
	while(FOREVER) {
		if (next==NULL) break;
		root=next;
		next=root->next;		// Got to find next before we delete
		destroy_node(root);		// Delete the node and all descendants
		if (next==min) break;	// If the next node is the minimum node, we are done
	}
#endif

}
/******************************************************************************/
void FibonacciHeap::clear(void)
{
	destroyData();
	min=NULL;
	max_rank=0;
	nroot=0;
	nnode=0;
}
/******************************************************************************/
void FibonacciHeap::destroyData(void)
{
	destroyData(min);
}
/******************************************************************************/
void FibonacciHeap::destroyData(FNODE *start)
{
	FNODE *next,*node;

	next=start;
	while(FOREVER) {
		if (next==NULL) break;
		node=next;
		next=node->next;
		delete node->item;
		node->item=NULL;
		destroyData(node->child);
		if (next==start) break;
	}
}
/******************************************************************************/
// Delete a nodes and all of the descendants
// Does not get called if we preallocated nodes
void FibonacciHeap::destroy_node(FNODE *node)
{
	FNODE *child,*next;

	// If node has children
	if (node->child!=NULL) {

		// Destroy all of the children first
		next=node->child;
		while(FOREVER) {
			if (next==NULL) continue;
			child=next;
			next=child->next;
			destroy_node(child);
			if (next==node->child) break;	// If the next is the start, we are done
		}
	}
	if ((node->up)!=NULL) ((node->up)->rank)--;
	
	delete node;
	nnode--;
}
/******************************************************************************/
// This merges the chain starting at node with the heap. This assumes the min is 
// in this node
void FibonacciHeap::meld(FNODE *node)
{
	int n;
	FNODE *tmp;

	// If no min make this the min
	if (min==NULL) {
		min=node;
		min->prev=min;
		min->next=min;
	}
	else {
		n=compare_nodes(min,node);

		// If node is smaller, swap min and node 
		if (n==2) {
			tmp=min;
			min=node;
			node=tmp;

			// Cut old min (now node) out of root chain and put
			// the new min in the chain
			if (node->next!=node) {
				// Set the neighbors to point to the new min
				(node->prev)->next=min;
				(node->next)->prev=min;

				// Set the new min to point to the neighbors
				min->next=node->next;
				min->prev=node->prev;

				// Cut the old min out
				node->prev=NULL;
				node->next=NULL;
			}
			else {	// Only one
				min->prev=min;
				min->next=min;
			}
			min->is_root=TRUE;
		}

		// Add the larger to the smaller as a child
		add_child(min,node);
	}

}
/******************************************************************************/
void FibonacciHeap::add_child(FNODE *parent, FNODE *node)
{
	FNODE *first_child;

	// If no children - make this the only child
	first_child=parent->child;
	if (first_child==NULL) {
		parent->child=node;
		node->next=node;		// Circular chain points to itself since only one
		node->prev=node;
	}
	else { // Else add previous to the child
		node->prev=first_child->prev;	// Point to our left sibling
		node->next=first_child;			// Point to our right sibling

		(first_child->prev)->next=node;	// Set our left sibling to point to us	(next)
		first_child->prev=node;			// Set our right sibling to point to us (prev)
	}
	node->up=parent;			// Set our parent
	parent->rank++;
	node->is_root=FALSE;

	if (parent->rank>max_rank) max_rank=parent->rank;
}
/******************************************************************************/
FNODE *FibonacciHeap::create_node(void)
{
	FNODE *node;

#ifdef PREALLOCATE
	// Allocate memory from the global preallocated memory
	node=&(chunk[myChunk].global_fnode[chunk[myChunk].nglobal_fnode]);
	chunk[myChunk].nglobal_fnode++;
	assert(chunk[myChunk].nglobal_fnode<NPRENODES);
#else
	node=new FNODE;
#endif
	if (node!=NULL) {
		node->up=NULL;
		node->child=NULL;
		node->item=NULL;
		node->key=DBL_MAX;
		node->marked=FALSE;
		node->is_root=FALSE;
		node->rank=0;
		node->next=NULL;
		node->prev=NULL;
	}
	nnode++;
	return(node);
}
/******************************************************************************/
void *FibonacciHeap::get_min(void)
{
	if (min!=NULL) return(min->item);
	else return(NULL);
}
/******************************************************************************/
FNODE *FibonacciHeap::insert_item(void *item, double key)
{
	FNODE *node;
	node=create_node();
	if (node!=NULL) {
		node->item=item;
		node->key=key;
		meld(node);
	}
	return(node);
}
/******************************************************************************/
void *FibonacciHeap::delete_min(void)
{
	int i;
	FNODE *node,*start;
	PFNODE *array,*list;
	int rank,done=FALSE,nlist,nrank,c;
	void *old_min_item;
	int old_min_rank;
	int old_max_rank;

	if (min==NULL) return(NULL);

	// Theoretically all children could be linked which would add (min->rank) to max_rank
	nrank=(max_rank+min->rank)*2;
	nrank=nnode;
	old_min_rank=min->rank;
	old_max_rank=max_rank;

	// Allocate array to hold various ranks
	assert(nrank<NPRENODES);
	array=global_arry;

	if (array==NULL) {
		put_error("Error:No memory in FibonacciHeap::delete_min");
		return(NULL);
	}
	for (i=0;i<nrank;i++) array[i]=NULL;	// Init array

	// Allocate array to hold all children
	assert(min->rank+nroot<NPRENODES);
	list=global_list;

	// Save all the children of min in the list
	nlist=0;
	node=min->child;
	while(FOREVER) {
		if (node==NULL) break;
		list[nlist]=node;
		node->up=NULL;		// We have no more parent
		nlist++;
		node=node->next;
		if (node==min->child) break;
	}

	// Save all the roots in the list
	node=min->next;
	while(FOREVER) {
		if (node==min) break;
		list[nlist]=node;
		nlist++;
		node=node->next;
	}

	// Isolate the nodes in list
	for (i=0;i<nlist;i++) {
		list[i]->next=NULL;
		list[i]->prev=NULL;
	}

	// Free the old min
	if (min->next!=min) {
		(min->next)->prev=min->prev;
		(min->prev)->next=min->next;
	}
	old_min_item=min->item;

	// Old way
	//delete min;
	nnode--;
	min=NULL;

	// Merge all roots in lists with the same rank
	nroot=nlist;	// Number of new possible roots
	for (i=0;i<nlist;i++) {

		// Link the roots forever until we find a unique rank
		node=list[i];
		while(FOREVER) {
			rank=list[i]->rank;

			if (array[rank]!=NULL) {
				node=link(node,array[rank]);
				array[rank]=NULL;
				nroot--;	// Got rid of one possible root
			}
			else {
				array[rank]=node;
				break;
			}
		}
	}

	// Now create circular link for the remaining
	start=NULL;
	for (i=0;i<nrank;i++) {

		// If something is in this spot in the array
		if (array[i]!=NULL) {
			array[i]->is_root=TRUE;	// Make it a root
			if (start==NULL) {		// If this is the first one
				start=array[i];		// Set start
				min=array[i];		// Make it the minimum
				min->prev=min;
				min->next=min;
			}
			else {					
				// Add to the root chain
				array[i]->next=min;
				array[i]->prev=min->prev;
				(min->prev)->next=array[i];
				min->prev=array[i];

				// Check if this node is less than the current min
				c=compare_nodes(min,array[i]);
				if (c==2) min=array[i];	// We have a new min
			}
		}
	}

	//delete [] array;
	//delete [] list;

	return(old_min_item);
}
/******************************************************************************/
int FibonacciHeap::compare_nodes(FNODE *node1, FNODE *node2)
{
	if (node1==NULL) return(1);
	if (node2==NULL) return(2);
	if (node1->key<node2->key) return(1);
	else return(2);
}
/****************************************************************************/
FNODE *FibonacciHeap::link(FNODE *node1, FNODE *node2)
{
	int n;

	n=compare_nodes(node1,node2);
	// Add the larger to the smaller as a child
	if (n==1) {
		add_child(node1,node2);
		return(node1);
	}
	else {
		add_child(node2,node1);
		return(node2);
	}
}
/******************************************************************************/
void FibonacciHeap::decrease_key(FNODE *node, double new_key)
{
	// Decrease key
	node->key=new_key;

	// If this node is the min we are fine
	if (min==node) return;

	// Cut the links to the parent and siblings
	if (node->prev==node) {		// If we are an only child
		if ((node->up)!=NULL) {
			((node->up)->rank)--;	// Decrease the rank of the parent
			(node->up)->child=NULL; // No favorite child
		}
	}
	else {
		// Set our prev brother to point at our next brother
		(node->prev)->next=node->next;

		// Set our next brother to point at our prev brother
		(node->next)->prev=node->prev;

		// If we have a parent
		if ((node->up)!=NULL) {
			// If we were the favorite child switch to our next sibling
			if ((node->up)->child==node) (node->up)->child=node->next;

			((node->up)->rank)--;	// Decrease the rank of the parent
		}
	}
	node->up=NULL;		// Either way we now have no parent

	// Add node to root list
	add_root(node);

	// If this key is now less than the min make it the min
	if (node->key<min->key) min=node;
}
/******************************************************************************/
void FibonacciHeap::add_root(FNODE *node)
{
	node->up=NULL;			// No more parent
	node->prev=min->prev;
	node->next=min;
	(min->prev)->next=node;
	min->prev=node;
	node->is_root=TRUE;
	nroot++;
}
/******************************************************************************/
void *FibonacciHeap::delete_node(FNODE *node)
{
	FNODE *parent,*child,*next;
	void *old_item;

	// If node is the minimum do a delete min instead
	if (node==min) {
		return(delete_min());
	}
	// If node is a root cut from the list of roots
	else if (node->is_root) {
		// Cut from list
		(node->next)->prev=node->prev;
		(node->prev)->next=node->next;
		nroot--;
	}
	// Not a root cut from the parent
	else {	
		parent=node->up;
		if (parent!=NULL) parent->rank--;
		if (node->next==node) {
			if (parent!=NULL) parent->child=NULL;
		}
		else {
			if (node==parent->child) {
				if (parent!=NULL) parent->child=node->next;
			}
			(node->next)->prev=node->prev;
			(node->prev)->next=node->next;
		}
	}

	// Add node's children to the root list
	child=node->child;
	while(FOREVER) {
		if (child==NULL) break;
		next=child->next;
		add_root(child);
		child=next;
		if (child==node->child) break;	// Looped to the start
	}

	old_item=node->item;
	// Old way
	// delete node;
	node=NULL;
	return(old_item);
}
/******************************************************************************/
void FibonacciHeap::check_heap(void)
{
	int max=0;
	FNODE *root,*next;

	// Destroy all root nodes
	next=min;
	while(FOREVER) {
		if (next==NULL) break;
		root=next;
		next=root->next;		// Got to find next before we delete
		max+=check_node(root);	// Delete the node and all descendants
		if (next==min) break;	// If the next node is the minimum node, we are done
	}
}
/******************************************************************************/
// Delete a nodes and all of the descendants
int FibonacciHeap::check_node(FNODE *node)
{
	FNODE *child,*next;
	int max=0;

	if (node->child!=NULL) {

		// Destroy all of the children first
		next=node->child;
		while(FOREVER) {
			if (next==NULL) continue;
			child=next;
			next=child->next;
			max+=check_node(child);
			if (next==node->child) break;	// If the next is the start, we are done
		}
	}
	max+=node->rank;
	return(max);
}
/******************************************************************************/
