/***************************************************************************/
/*
	Graph.h
	Written by Matthew Maxel
	Created Dec 28, 2001

	The file defines a class to encapsulate the graph and its structures and 
	do internal manipulation of the graph.
*/
/******************************************************************************/
#include "BranchingStructureNode.h"
#include "edge.h"
#include "Path.h"				// Uses Path
#include "NodeLabels.h"
#include "FibonacciHeap.h"
#include "edgeKey.h"
#include "DataParam.h"
#include "Node.h"
/***************************************************************************/
#ifndef Graph_h
#define Graph_h
/***************************************************************************/



class Graph 
{
	// Generated data
	DataParameters p0;			// List of parameters used to generate data
	Node *nodes;				// Pointer to list of points
	Edge *edges;				// Pointer to list of edges sorted by source
	int startNode;				// Start node for graph
	int endNode;				// End node for graph
	int hybridNumber;			// Number for which it is cheaper to do repeated Dykstras

	// Computed data that we keep track of
	int *forwardIndex;			// Index to edges sorted by source node
	EdgeList *forwardEdgeList;	// New edge list 
	int *reverseIndex;			// Index to edges sorted by destination node
	EdgeList *reverseEdgeList;	// New destination edge list 

	NodeLabels *labels;			// Node Labels for forward and reverse info
	NodeLabels *scratchLabels;	// Scratch labels for subgraphs
	NodeLabels *scratchLabels2;	// Scratch labels for naive shortest path  
	Edge *scratchEdges;			// Scratch edges list
	Edge *reverseEdges;			// Reverse of all edges
	EdgeKey *edgeKey;			// Edge keys

	Path shortestPath;			// Shortest path

	// Stats
	int numBadCases;			// Number of bad cases
	int numBranchPaths;			// Number of branch paths created
	int totalBranchLengths;		// Total length of those paths
	int numKSP;					
	int totalKSPLengths;		//
	int numDijkstras;			// Number of Dijkstra's run
	int numNodesScanned;		// Total nodes scanned in those Dijkstra's

	int ordered;				// Graph has been ordered in a semicircle
	int labelsValid;			// Labels are valid

	int checkForNegativeCycle(NodeLabels *labels, Edge *edges,int nedges);

public:
	Graph(void);
	~Graph(void);
        Graph(Graph &p);

	void freeAll(void);					// Frees data structures and deletes graphics
	int getStartNode(void);				// Gets the start node of the graph
	void setStartNode(int n);			// Sets the	start node of the graph
	int getEndNode(void);				// Gets the end node of the graph
	void setEndNode(int n);				// Sets the end node of the graph
	DataParameters getParameters(void);
	void setParameters(DataParameters p);

	void generate(DataParameters p);			// Generate a graph with parameters p
	void randomGenerate(DataParameters param);	// Randomly generate a graph
	void gridGenerate(DataParameters p);		// Generate points on a grid

	void computeEdgeLists(void);				// Computes the edges lists from the edges
	int order(void);	// Orders the graph so the shortest path is a 
						// straight edge and all other points are above it
	int input(char *path);
	int inputRaw(char *path);
	int output(char *path);

	void sortPoints(int (*cmp)(void *, void *));	// Sorts the points 
	int checkPtsExist(void);			// returns whether points exist or not

	void createDrawingForPath(const Path& path, char *name, int color, double dx, double dy);
	void createDrawingForLabels(void);
	void createDrawingForPts(void);
	void createDrawingForPt(int n);
	void createDrawing(void);
	void createDrawingForShortestPath(void);
	void createDrawingForEdge(int n, int color, int linetype, double x_offset, double y_offset, int text_flag);
	void createDrawingForEdge(Edge &e, int color, int linetype, double x_offset, double y_offset, int text_flag);
	long mapColor(int n);

	Edge getEdge(int n);
	Edge getReverseEdge(int n);
	void setEdge(int n, Edge edgeIn);			// Set edge n to edgeIn
	int findEdge(int p1, int p2);	// Finds an edge with source node p1 and destination node p2
	void setEdges(Edge *edges);					// Set the array of edges

	int solveReverseShortestPath(void);
	int computeReverseInfo(DataParameters p, NodeLabels *label, int endNode0);
	int getNpath(void);					// Returns number of hops in sp
	int getPath(int n);					// Returns path # of node n
	int getBlock(int n);				// Returns block # of node n
	int getMinBlock(int n);				// Returns minblock # of node n
	double getCost(int n);				// Returns cost of node n
	double getCostToY(int n);			// Return cost of node n to y
	void getShortestPath(Path &path);	// Return shortest path
	double getPtVal(int n, int d);		// Returns the x or y position of point n

	EdgeList getEdgeList(int node);		// Gets a list of edges for a node
	EdgeList *getEdgeList(void);		// Gets a pointer to the whole edge list

	NodeLabels *getLabels(void);
	int computeBlock(void);				// Computes block and minblock
	
	// Set functions - be careful
	void setLabels(NodeLabels *labelsIn);
	void deleteDrawings(void);
	void checkForDuplicateEdge(void);

	int *getForwardIndex(void);
	void setForwardIndex(int *forwardIndexIn);
	int *getReverseIndex(void);
	void setReverseIndex(int *reverseIndexIn);
	void setForwardEdgeList(EdgeList *forwardEdgeListIn);
	void setReverseEdgeList(EdgeList *reverseEdgeListIn);
	void setNodes(Node *nodesIn);
	EdgeList *getForwardEdgeList(void);
	EdgeList *getReverseEdgeList(void);

	// Computes path for replacement edge
	int computePath(int e, Path &path);

	// Computes shortest path from labels
	int computePath(NodeLabels *labels, int endNode0, Path &path);

	int checkPath(const Path &path);
	double calculatePathCost(const Path &path);
	void formatPath(const Path& path, char *msg);
	void formatPath(char *pathName, const Path &path, char *msg);
	void logPath(const Path& path);
	
	// Routines to solve the shortest path
	// These routines solve shortest path for our graph only
	int solveShortestPath(int method);			// Method 0 - fastest, 1 - complete
	int solveShortestPathDijkstra(int method);
	int solveShortestPathBellmanFord(void);

	int solveBellmanFord(void);	// Computes node labels with Bellman-Ford

	// These routine solve shortest path for any graph
	int solveBellmanFord(int startNode0, int endNode0, int npts0, 
		Edge *edges0, int nedges0, NodeLabels *labels);

	void copy(Graph *H);
	void release(void);
	void setScratchLabels(NodeLabels *scratchIn);
	void setScratchLabels2(NodeLabels *scratchIn);
	void setNodeInvalid(int n);
	void setEdgeInvalid(int e);
	void setReverseEdges(Edge *reverseEdgesIn);
	int findReplacementPaths(Path **paths_out, int *npath_out);
	void setEdgeKey(EdgeKey *edgeKeyIn);
	int findBestReplacementPath(int len, Path &path);

	void genSpace(DataParameters param);
	void precomputation(DataParameters param);
	int oldFindReplacementPaths(Path **paths_out, int *npath_out);

	
	int getNumBadCases(void);
	void resetNumBadCases(void);
	int getNumDijkstras(void);
	void resetNumDijkstras(void);
	int getNumBranchPaths(void);
	void resetNumBranchPaths(void);
	int getTotalBranchLengths(void);
	void resetTotalBranchLengths(void);
	int getNumKSP(void);
	void resetNumKSP(void);
	int getTotalKSPLengths(void);
	void resetTotalKSPLengths(void);
	int getNumNodesScanned(void);
	void resetNumNodesScanned(void);

	void find2BestReplacementPaths(Path *prefix, FibonacciHeap *h, BRNODE *u, BRNODE *v);

	// Return number of nodes in the graph
	int getNumNodes(void);

	// Solves the shortest path and replacement path for c(u) and c(u,v) respectively
	int findBestUnionClassPaths(BRNODE *u, BRNODE *v, FibonacciHeap *h, 
		const double prefixCost);

	// Computes as many as k paths from this graph
	int computeKSP(int k, Path *pathArray);

	// Computes the shortest path for a node class c(u)
	int computeNodeClassShortestPath(BRNODE *u, FibonacciHeap *h);

	// Computes the shortest path for a path class c(u,v)
	int computePathClassShortestPath(BRNODE *u, BRNODE *v, FibonacciHeap *h);

	// Computes the shortest paths for c(u) and c(u,v)
	int computeUnionClassShortestPath(BRNODE *u, BRNODE *v, FibonacciHeap *h);

	// Computes the shortest path for c(u,w) and c(w,v)
	int compute2PathClassShortestPath(BRNODE *u, BRNODE *w, BRNODE *v, 
		FibonacciHeap *h);

	// Add a path and its branching structure nodes to the heap
	void addPathToHeap(FibonacciHeap *h, const Path& path, BRNODE *u, BRNODE *v);

	// Translate an edge path to a node path
	void edgeToNode(const Path &p, Path& p2);

	int solveDijkstra(int method);			// Compute labels for shortest path for this graph
	int solveDijkstra(int method, NodeLabels *specificLabels); // Compute a specific set of labels
	int solveReverseDijkstra(int method);	// Compute labels (for y) for reverse shortest path 

	int solveNaiveReplacement(int e, Path &naivePath);


	int computeYenKSP(int K, Path *pathArray);
	int computeYenNextPath(FibonacciHeap *h, int &k, Path *pathsOut);
	void addYenPathToHeap(FibonacciHeap *h, const Path& pathIn);

	// New modification
	int computeNewKSP(int k, Path *pathArray);
	int computeNextShortestPath(BRNODE *t, FibonacciHeap *h);

	void setHybridNumber(int n);
};
 int  fastCompare(const void *p1, const void *p2);
 int fastRCompare(const void *p1, const void *p2);
 int fastCostCompare(const void *p1, const void *p2);
/***************************************************************************/
#endif
/***************************************************************************/
