/******************************************************************************/
/*
	GraphKSP.cpp
	Created Oct, 2001
*/
/******************************************************************************/
#include "Graph.h"
#include "BranchingStructure.h"
#include "const.h"
#include "ret.h"
#include "xrout.h"
#include "xmenus.h"
#include<stdio.h>
/******************************************************************************/
// Returns the actual number of paths found
int Graph::computeKSP(int k, Path *pathArray)
{
	int i;
	FibonacciHeap h;
	BranchingStructure T;
	int n,numPathsFound,bunch;
	BRNODE *u,*tp,*w;
	char msg[128];
	PATH_INFO *p;
	int showInfo=Sys.getShowInfo();

	// Catch invalid situations
	if (k<=0 || p0.npts==0) return(NO_CHANGE);

	// This computes the shortest path and other needed information
	solveShortestPath(1);
	if (shortestPath.nList==0) {
		//put_error("Error:No shortest path found");
		return(ZERO);
	}

	// Create first node in branching structure
	u=T.createRootNode(startNode);		

	// Add this path to the heap
	addPathToHeap(&h,shortestPath,u,NULL);		

	// Compute the k shortest paths
	numPathsFound=0;

	// Init stats
	numBadCases=0;
	numBranchPaths=0;
	totalBranchLengths=0;
	numDijkstras=0;
	numNodesScanned=0;

	bunch=k/20;
	if (bunch<1) bunch=1;
	for (i=0;i<k;i++) { 
		// Debug
		if (Sys.getDebug() && Sys.getLogging()) {
			Sys.logString("********************************************************************************\n");
			sprintf(msg,"Computing path %1d of %1d\n",i,k);
			Sys.logString(msg);
		}

		// 1. Extract the path associated with the minimum key from the heap
		p=(PATH_INFO *)(h.delete_min());
		if (p==NULL) break;	// If no more path break;
		if (p->path->nList==0) { 	// Check if we inserted a bad path into the heap
			//put_error("Error:Bad path found in heap");
			Sys.stopLogging();
			break;
		}

		// 2. If this path belongs to an equivalance class C(u)
		if (p->v==NULL) {

			// p.node1 is the node where this path starts
			// (a) Add a new branch (u,tp) to T that represents the suffix of P after u
			tp=T.addLeaf(this,p->u,p->path);
			if (tp==NULL) return(NOT_FOUND);

			// (b) Remove from C(u) the paths that share at least one edge with P after u and
			//	   put all of them except P into the newly created equivalence class C(u,tp).
			// Conceptual

			// For each new or changed class compute the shortest path form s to t that belongs
			// to the class and insert these path into the heap

			// If there is only one edge in between u and v then there is no new paths using
			// the replacement edge routine c(u,v), find the new c(u) paths the old way
			if (tp->prefixPath->nList==1) {
				computeNodeClassShortestPath(p->u,&h);
			}
			else { 
				if (tp->prefixPath->nList<=hybridNumber) {
					computeNodeClassShortestPath(p->u,&h);
					computePathClassShortestPath(p->u,tp,&h);
					numBranchPaths++;
					totalBranchLengths+=(p->path->nList);
				}
				else {
					computeUnionClassShortestPath(p->u,tp,&h);
					numBranchPaths++;
					totalBranchLengths+=(p->path->nList);
				}
			}
		}
		// 3. Else (P belongs to the equivanlence class C(u,v) for some branch (u,v)
		else {

			// Update the branching structure with a new branch and a new leaf
			w=T.addBranch(this,p->u,p->v,p->path);
			if (w==NULL) break;

			// Add a new leaf for the new branch
			tp=T.addLeaf(this,w,p->path);
			if (tp==NULL) break;

			// If there is only one edge in between u and v then there is no new paths using
			// the replacement edge routine c(u,v), find the new c(u) paths the old way
			if (tp->prefixPath->nList==1) {
				computeNodeClassShortestPath(w,&h);
			}
			else {
				computeUnionClassShortestPath(w,tp,&h);
				numBranchPaths++;
				totalBranchLengths+=(p->path->nList);
			}

			// Compute C(u,w) and C(w,v);
			if (w->prefixPath->nList>1 && p->v->prefixPath->nList>1) {
				compute2PathClassShortestPath(p->u,w,p->v,&h);
				numBranchPaths+=2;
				totalBranchLengths+=(w->prefixPath->nList+p->v->prefixPath->nList);
			}
			// New shortest path for C(u,w)
			else if (w->prefixPath->nList>1) {
				computePathClassShortestPath(p->u,w,&h);
				numBranchPaths++;
				totalBranchLengths+=(w->prefixPath->nList);
			}
			// New shortest path for C(w,v)
			else if (p->v->prefixPath->nList>1) {
				computePathClassShortestPath(w,p->v,&h);
				numBranchPaths++;
				totalBranchLengths+=(p->v->prefixPath->nList);
			}
		}

		// Report status to the user periodically
		if (showInfo) {
			if ((i%bunch)==0) {
				//sprintf(msg,"Computing Path:%3d",i);
				//show_msg(msg);
			}
		}
	}

	T.computePaths(this,pathArray,&numPathsFound);

	// Compute stats
	numKSP=numPathsFound;
	totalKSPLengths=0;
	for (i=0;i<numPathsFound;i++) 
		totalKSPLengths+=pathArray[i].nList;

	// If debugging check the paths and show number of bad cases
	if (showInfo && Sys.getDebug())	{
		for (i=0;i<numPathsFound;i++) {
			n=checkPath(pathArray[i]);
		}
		sprintf(msg,"Number of bad cases %3d",numBadCases);
		show_side_msg(msg);
	}

	// Delete the paths off the heap before the heap is destroyed
	h.destroyData();

	return(numPathsFound);
}
/***************************************************************************/
// Computes graph H from G and calls findBestUnionClassPaths to find the 
// shortest	path in C(u) and C(u,u) 
int Graph::computeUnionClassShortestPath(BRNODE *u, BRNODE *v, FibonacciHeap *h)
{
	int i;
	Edge edge;
	BRNODE *node;
	Graph H;
	BRLIST::iterator it;
	double prefixCost=0.0;

	// Implementation
	// Fast copy G into H 
	// Delete (make infinity) edges we don't want
	// Delete (make invalid) nodes we don't want

	copy(&H);	// Copy structures to H

	// Delete all nodes on the total prefix path of u
	// Also calculate the cost of getting to u
	node=u;
	while(FOREVER) {
		if (node==NULL) break;
		for (i=0;i<node->prefixPath->nList;i++) {
			edge=H.getEdge(node->prefixPath->list[i]);
			H.setNodeInvalid(edge.p1);
		}
		prefixCost+=node->prefixPath->cost;
		node=node->up;
	}

	// Start at node represented by u
	H.setStartNode(u->node);

	// Delete all known paths from node u except our edge
	for (it=u->children.begin();it!=u->children.end();++it) {
		node=(*it);
		if (node!=v) 
			H.setEdgeInvalid(node->prefixPath->list[0]);
	}

	// This will compute the best replacement paths for the node and 
	// the path from u to v
	H.findBestUnionClassPaths(u,v,h,prefixCost);

	// Add the number of bad cases to the global number  
	numDijkstras+=H.getNumDijkstras();
	numNodesScanned+=H.getNumNodesScanned();
	numBadCases+=H.getNumBadCases();

	// Set these parameters to NULL so H does not delete them when it destructs
	H.release();	// Release computed information that belongs to G

	// When H dies it should kill our labels as well
	return(OK);
}
/***************************************************************************/
// Creates a new instance of the next shortest path for a class of paths
int Graph::compute2PathClassShortestPath(BRNODE *u, BRNODE *w, BRNODE *v, 
	FibonacciHeap *h)
{
	int i;
	Edge edge;
	BRNODE *node;
	int b;
	Path pathOut,newPath;
	Graph H;
	double cost;

	// Copy G into H 
	// Delete (make infinity) edges we don't want
	// Delete (make invalid) nodes we don't want
	copy(&H);	// Copy structures to H

	//-------------------------------------------------------------------------------
	// Part 1
	// Delete all nodes on the total prefix path of u
	node=u;
	cost=0.0;
	while(FOREVER) {
		if (node==NULL) break;

		// This node's partial path
		for (i=0;i<node->prefixPath->nList;i++) {
			H.setNodeInvalid(edges[node->prefixPath->list[i]].p1);
		}

		cost+=node->prefixPath->cost;
		node=node->up;
	}

	// Find b 
	// b is from the lead of path(a,c)
	edge=edges[w->prefixPath->list[0]];
	b=edge.p2;	
	H.setStartNode(b); // Start at b for part 1

	// Delete node u
	H.setNodeInvalid(u->node); 

	// If the path is less than 3 more effiscient to do 3 shortest paths
	if (w->prefixPath->nList<=hybridNumber) {
		double minCost=INF;
		for (i=1;i<w->prefixPath->nList;i++) {
			H.solveNaiveReplacement(w->prefixPath->list[i],pathOut);
			if (pathOut.cost<minCost) {
				newPath=pathOut;
				minCost=newPath.cost;
			}
		}
	}
	else {
		// Get the best replacement path upto w->prefixPath->nList-1
		H.findBestReplacementPath(w->prefixPath->nList,newPath);
	}
	if (newPath.nList) {
		newPath.pushFront(w->prefixPath->list[0],edge.cost,pathOut);
		pathOut.cost+=cost;	// Tack on the cost of the prefixpath
		addPathToHeap(h,pathOut,u,w);
	}

	//-------------------------------------------------------------------------------
	// Part 2
	// Delete all nodes on the total prefix path of w
	newPath.freeList();	// Reset newPath
	node=w;
	//cost=0.0; // cost is cost from startNode to u plus u to w
	while(FOREVER) {
		if (node==NULL) break;

		// This node's partial path
		for (i=0;i<node->prefixPath->nList;i++) 
			H.setNodeInvalid(edges[node->prefixPath->list[i]].p1);

		cost+=node->prefixPath->cost;
		node=node->up;
		if (node==u) break;	// We already did node u and up
	}

	// Find b 
	// b is from the lead of path(a,c)
	edge=edges[v->prefixPath->list[0]];
	b=edge.p2;	
	H.setStartNode(b); // Start at b for part 1

	// Delete node w
	H.setNodeInvalid(w->node); 

	// If the path is less than 3 more effiscient to do 3 shortest paths
	if (v->prefixPath->nList<=hybridNumber) {
		double minCost=INF;
		for (i=1;i<v->prefixPath->nList;i++) {
			H.solveNaiveReplacement(v->prefixPath->list[i],pathOut);
			if (pathOut.cost<minCost) {
				newPath=pathOut;
				minCost=newPath.cost;
			}
		}
	}
	else {
		// Get the best replacement path upto v->prefixPath->nList-1
		H.findBestReplacementPath(v->prefixPath->nList,newPath);
	}
	if (newPath.nList) {
		newPath.pushFront(v->prefixPath->list[0],edge.cost,pathOut);
		pathOut.cost+=cost;	// Tack on the cost of the prefixpath
		addPathToHeap(h,pathOut,w,v);
	}

	//-------------------------------------------------------------------------------
	// Set these parameters to NULL so H does not delete them when it destructs
	numDijkstras+=H.getNumDijkstras();
	numNodesScanned+=H.getNumNodesScanned();
	numBadCases+=H.getNumBadCases();
	H.release();			// Release computed information because it really belongs to G

	// When H dies it should kill our labels as well
	return(OK);
}
/***************************************************************************/
// Creates a new instance of the next shortest path for a class of paths
int Graph::computeNodeClassShortestPath(BRNODE *u, FibonacciHeap *h)
{
	int i;
	BRNODE *node;
	Path pathOut;
	Graph H;
	BRLIST::iterator it;
	double cost;

	// Copy G into H 
	// Delete (make infinity) edges we don't want
	// Delete (make invalid) nodes we don't want
	copy(&H);	// Copy structures to H

	// Delete all nodes on the total prefix path of u
	node=u;
	cost=0.0;
	while(FOREVER) {
		if (node==NULL) break;
		for (i=0;i<node->prefixPath->nList;i++) 
			H.setNodeInvalid(edges[node->prefixPath->list[i]].p1);
		cost+=node->prefixPath->cost;
		node=node->up;
	}

	// Start at node represented by u
	H.setStartNode(u->node);

	// Delete all known paths from node u
	for (it=u->children.begin();it!=u->children.end(); ++it) {
		node=(*it);
		H.setEdgeInvalid(node->prefixPath->list[0]);
	}

	// Solve for the shortest path
	H.solveShortestPath(0);
	H.getShortestPath(pathOut);

	// Set these parameters to NULL so H does not delete them when it destructs
	numDijkstras+=H.getNumDijkstras();
	numNodesScanned+=H.getNumNodesScanned();
	numBadCases+=H.getNumBadCases();
	H.release();			// Release computed information because it really belongs to G

	if (pathOut.nList>0) {
		pathOut.cost+=cost;	// Tack on the cost of the prefixpath
		addPathToHeap(h,pathOut,u,NULL);
	}

	// When H dies it should kill our labels as well
	return(OK);
}
/***************************************************************************/
// Creates a new instance of the next shortest path for a class of paths
int Graph::computePathClassShortestPath(BRNODE *u, BRNODE *v, FibonacciHeap *h)
{
	int i,b;
	Edge edge;
	BRNODE *node;
	Path pathOut,newPath;
	Graph H;
	BRLIST::iterator it;
	double cost;

	// u and v must be separate by 1 or more nodes
	if (v->prefixPath->nList<=1) return(NULL);

	// Basically copy G into H 
	// Delete (make infinity) edges we don't want
	// Delete (make invalid) nodes we don't want
	copy(&H);	// Copy structures to H

	// Delete all nodes on the total prefix path of u
	node=u;
	cost=0.0;
	while(FOREVER) {
		if (node==NULL) break;
		for (i=0;i<node->prefixPath->nList;i++) {
			edge=H.getEdge(node->prefixPath->list[i]);
			H.setNodeInvalid(edge.p1);
		}
		cost+=node->prefixPath->cost;
		node=node->up;
	}

	// Find b 
	// b is from the lead of path(a,c)
	edge=edges[v->prefixPath->list[0]];
	b=edge.p2;	

	// Start at b 
	H.setStartNode(b);

	// Delete node u
	H.setNodeInvalid(u->node); 

	// If the path is less than 3 more effiscient to do 3 shortest paths
	if (v->prefixPath->nList<=hybridNumber) {
		double minCost=INF;
		for (i=1;i<v->prefixPath->nList;i++) {
			H.solveNaiveReplacement(v->prefixPath->list[i],pathOut);
			if (pathOut.cost<minCost) {
				newPath=pathOut;
				minCost=newPath.cost;
			}
		}
	}
	else {
		// Get the best replacement path upto v->prefixPath->nList-1
		H.findBestReplacementPath(v->prefixPath->nList,newPath);
	}

	if (newPath.nList) {
		newPath.pushFront(v->prefixPath->list[0],edge.cost,pathOut);
		pathOut.cost+=cost;	// Tack on the cost of the prefixpath
		addPathToHeap(h,pathOut,u,v);
	}

	// Set these parameters to NULL so H does not delete them when it destructs
	numDijkstras+=H.getNumDijkstras();
	numNodesScanned+=H.getNumNodesScanned();
	numBadCases+=H.getNumBadCases();
	H.release();			// Release computed information because it really belongs to G

	// When H dies it should kill our labels as well
	return(OK);
}
/***************************************************************************/
void Graph::addPathToHeap(FibonacciHeap *h, const Path& path, BRNODE *u, BRNODE *v)
{
	PATH_INFO *new_p=new PATH_INFO;

	new_p->path=new Path(path);
	new_p->u=u;
	new_p->v=v;

	if (Sys.getDebug() && Sys.getLogging()) {
		char msg[128];
		if (v==NULL) sprintf(msg,"Adding to heap c(%1d)",u->node);
		else sprintf(msg,"Adding to heap c(%1d,%1d)",u->node,v->node);
		Sys.logString(msg);
		logPath(path);
	}

	h->insert_item(new_p,new_p->path->cost);
}
/***************************************************************************/
