/******************************************************************************/
/*
	GraphNewKSP.cpp
	Created Oct, 2001
*/
/******************************************************************************/
#include <math.h>
#include <fstream>
#include <string>
#include <sstream>
#include <assert.h>
#include <cstdlib>
#include <map>
/******************************************************************************/
#include "Graph.h"
#include "BranchingStructure.h"
#include "FibonacciHeap.h"
#include "NodeLabels.h"
#include "Path.h"
#include "LinkList.h"
#include "DataParam.h"
#include "Node.h"
/******************************************************************************/
#include "const.h"
#include "xstats.h"
#include "xmenus.h"
#include "xutil.h"
#include "xdrw.h"
#include "xgraphic.h"
#include "xprim.h"
#include "xcompare.h"
#include "colors.h"
#include "ret.h"
#include "xrout.h"
/******************************************************************************/
// Returns the actual number of paths found
int Graph::computeNewKSP(int k, Path *pathArray)
{
	int i;
	FibonacciHeap h;
	BranchingStructure T;
	int n,numPathsFound,bunch;
	BRNODE *s,*t;
	char msg[128];
	PATH_INFO *p;
	int showInfo=Sys.getShowInfo();
	int debugging=Sys.getDebug();
	int logging=Sys.getLogging();

	// Catch invalid situations
	if (k<=0 || p0.npts==0 || labels==NULL) return(NO_CHANGE);

	if (debugging && logging) {
		Sys.logString("********************************************************************************\n");
		sprintf(msg,"Computing path %1d of %1d\n",1,k);
		Sys.logString(msg);
	}

	// This computes the shortest path and other needed information
	solveShortestPath(1);
	if (shortestPath.nList==0) {
		//put_error("Error:No shortest path found");
		return(ZERO);
	}
	pathArray[0]=shortestPath;

	// Create first node in branching structure
	s=T.createRootNode(startNode);		

	numPathsFound=0;
	numBadCases=0;
	bunch=k/20;
	if (bunch<1) bunch=1;

	// Add the first path which is a unique case
	t=T.addLeaf(this,s,&shortestPath);
	computeNextShortestPath(t,&h);

	// Compute the k shortest paths
	for (i=1;i<k;i++) { 

		// Debug
		if (debugging && logging) {
			Sys.logString("********************************************************************************\n");
			//sprintf(msg,"Computing path %1d of %1d\n",i+1,k);
			Sys.logString(msg);
		}

		// 1. Extract the path associated with the minimum key from the heap
		p=(PATH_INFO *)(h.delete_min());
		if (p==NULL) break;	// If no more path break;
		if (p->path->nList==0) { 	// Check if we inserted a bad path into the heap
			put_error("Error:Bad path found in heap");
			Sys.stopLogging();
			break;
		}

		// Save the path in the return array
		pathArray[i]=(*(p->path));

		//  P always belongs to an equivalence class C(u,v) for some branch (u,v)
		// Update the branching structure with a new branch and a new leaf
		t=T.addNewBranch(this,p->u,p->v,p->path);
		if (t==NULL) break;

		// Compute best replacement in parent path
		computeNextShortestPath(p->v,&h);

		// Compute best replacement in new path
		computeNextShortestPath(t,&h);

		// Report status to the user periodically
		if (showInfo) {
			if ((i%bunch)==0) {
				//sprintf(msg,"Computing Path:%3d",i);
				//show_msg(msg);
			}
		}
	}

	//T.computeNewPaths(this,pathArray,&numPathsFound);
	numPathsFound=i;

	// If debugging check the paths and show number of bad cases
	if (showInfo && Sys.getDebug())	{
		for (i=0;i<numPathsFound;i++) {
			n=checkPath(pathArray[i]);
			if (n!=OK) {
				char msg[128];
				sprintf(msg,"Error:In path:%d",i);
				put_error(msg);
			}
		}
		sprintf(msg,"Number of bad cases %3d",numBadCases);
		//show_side_msg(msg);
	}

	// Delete the paths off the heap before the heap is destroyed
	h.destroyData();

	return(numPathsFound);
}
/***************************************************************************/
// Creates a new instance of the next shortest path for a class of paths
int Graph::computeNextShortestPath(BRNODE *t, FibonacciHeap *h)
{
	int i,j,c,e,npp,np,nBRNodes,nEdgesBeforeU,nEdgesAfterU;
	int firstPath,uFound,first;
	Path aPath,bestPath,path,prePath;
	Graph H;
	BRNODE *node,*u;
	BRLIST::iterator it,it0;
	PBRNODE *brlist;

	//assert(t!=NULL);

	// Store branching nodes in a list in forward order
	// 1. Compute number of branching nodes and
	// 2. Compute u - the top most node in the replacement path
	// 3. Compute number of nodes before u
	// 4. Compute number of nodes after u
	// 5. Compute whether we are firstPath or not
	c=0;
	nEdgesBeforeU=0;
	nEdgesAfterU=0;
	node=t;
	u=NULL;
	while(FOREVER) {
		c++;

		// Count how many nodes are on both sides of u
		if (u==NULL) nEdgesAfterU+=node->prefixPath->nList;
		else nEdgesBeforeU+=node->prefixPath->nList;

		// Figure out where u is and if this the original shortest path (firstPath)
		if (u==NULL) {
			it0=(node->up)->children.begin();
			if ((*it0)!=node) {
				u=node->up;
				firstPath=FALSE;
			}
			else {
				// If we are at the top this is our u then we are first path
				if ((node->up->up)==NULL) {
					u=node->up;
					firstPath=TRUE;
				}
			}
		}

		node=node->up;
		if (node->up==NULL) break;
	}
	nBRNodes=c+1;	// Include the node above the current node

	// If we are not first path add the lead edge to number of nodes in the prepath 
	// and subtract it from the number of nodes in the path
	if (firstPath==FALSE) {
		nEdgesBeforeU++;
		nEdgesAfterU--;
	}

	// Check if there is no way to have deviation
	if (nEdgesAfterU==0) 
		return(NO_CHANGE);

	// Basically copy G into H 
	// Delete (make infinity) edges we don't want
	// Delete (make invalid) nodes we don't want
	copy(&H);	// Copy structures to H

	// Now store branching nodes in a list in forward order
	c=0;
	brlist=new PBRNODE[nBRNodes];
	node=t;
	while(FOREVER) {
		if (node==NULL) break;
		brlist[nBRNodes-c-1]=node;
		c++;
		node=node->up;
	}

	// Set our start point
	// Create prePath and path
	// Delete prePath from H
	// Delete all known paths deviating from path
	uFound=FALSE;
	prePath.nList=nEdgesBeforeU;
	prePath.cost=0.0;
	prePath.list=new int[prePath.nList];
	path.nList=nEdgesAfterU;
	path.cost=0.0;
	path.list=new int[path.nList];

	npp=0;
	np=0;
	first=TRUE;
	for (i=0;i<nBRNodes;i++) {
		node=brlist[i];

		// Delete prePath nodes
		if (uFound==FALSE) {
			if (node!=u) H.setNodeInvalid(node->node);

			// Save prefix path
			for (j=0;j<node->prefixPath->nList;j++) {
				e=node->prefixPath->list[j];
				prePath.list[npp]=e;
				prePath.cost+=edges[e].cost;
				npp++;
				H.setNodeInvalid(edges[e].p1);
			}
		}
		else {
			// Save path
			for (j=0;j<node->prefixPath->nList;j++) {
				// If this is the first point after u and 
				// we are not the firstPath add the lead edge to prePath
				if (first && firstPath==FALSE) {
					prePath.list[npp]=node->prefixPath->list[j];
					prePath.cost+=edges[node->prefixPath->list[j]].cost;
					npp++;
					first=FALSE;
				}
				else {
					path.list[np]=node->prefixPath->list[j];
					path.cost+=edges[node->prefixPath->list[j]].cost;
					np++;
				}
			}

		}

		if (node==u) {
			uFound=TRUE;

			if (firstPath) H.setStartNode(u->node);
			else {
				e=brlist[i+1]->prefixPath->list[0];		// Lead edge
				H.setStartNode(edges[e].p2);    		// Start node
				H.setNodeInvalid(u->node);
			}
		}


		if (uFound==TRUE) {
			// Delete known deviations
			for (it=node->children.begin();it!=node->children.end();it++) {
				if ((*it)==brlist[i+1]) continue;
				else 
					H.setEdgeInvalid((*it)->prefixPath->list[0]);
			}
		}
	}

	// If the path is less than (hybridNumber=5) more effiscient to do Dijksta's
	if (path.nList<=hybridNumber) {
		double minCost=INF;
		for (i=0;i<path.nList;i++) {
			H.solveNaiveReplacement(path.list[i],aPath);
			if (aPath.cost<minCost) {
				bestPath=aPath;			// Could save only a pointer ? to save time
				minCost=bestPath.cost;
			}
		}
	}
	else {
		// Get the best replacement path upto t->prefixPath->nList-1
		H.findBestReplacementPath(path.nList+1,bestPath);
	}

	if (bestPath.nList) {
		// If we have a prepath add it back in
		if (prePath.nList) {
			aPath=prePath+bestPath;
			addPathToHeap(h,aPath,u,t);
		}
		else addPathToHeap(h,bestPath,u,t);
	}

	// Save the stats of H before its destroyed
	numDijkstras+=H.getNumDijkstras();
	numNodesScanned+=H.getNumNodesScanned();
	numBadCases+=H.getNumBadCases();

	// Set certain parameters to NULL so H does not delete them when it destructs
	// since they really belongs to G
	H.release();

	// When H dies it should kill our labels as well
	delete [] brlist;
	return(OK);
}
/***************************************************************************/
