/******************************************************************************/
/*
	Graph.cpp
	Created Oct, 2001
*/
/******************************************************************************/
#include <iostream>
#include <fstream>
#include <string.h>
#include <sstream>
#include <math.h>
#include <assert.h>
#include <cstdlib>
#include <map>
/******************************************************************************/
#include "Graph.h"
#include "BranchingStructure.h"
#include "FibonacciHeap.h"
#include "NodeLabels.h"
#include "Path.h"
#include "LinkList.h"
#include "DataParam.h"
#include "Node.h"
/******************************************************************************/
#include "const.h"
#include "xstats.h"
#include "xmenus.h"
#include "xutil.h"
#include "xdrw.h"
#include "xgraphic.h"
#include "xprim.h"
#include "xcompare.h"
#include "colors.h"
#include "ret.h"
#include "xrout.h"
/******************************************************************************/
using std::ofstream;
using std::ifstream;
/******************************************************************************/
int Graph::solveNaiveReplacement(int e0, Path &naivePath)
{
	int i,n;
	double saveCost;

	saveCost=edges[e0].cost;	// Save the cost
	edges[e0].cost=INF;			// Now make the cost infinite

	// Init the scratch labels
	for (i=0;i<p0.npts;i++) {
		scratchLabels2[i].valid=labels[i].valid;
		scratchLabels2[i].scanned=FALSE;
	}

	// Solve the shortest path problem with this new dataset
	if (p0.negativeEdgeCosts==FALSE) 
		n=solveDijkstra(0,scratchLabels2);
	else 
   		n=solveBellmanFord(
			startNode,
			endNode,
			p0.npts,
			edges,
			p0.nedges,
			labels);

	computePath(scratchLabels2,endNode,naivePath);	// Compute the path from the label info
	edges[e0].cost=saveCost;						// Restore the edge costs

	// Return whether we could solve problem or not
	return(OK);
}
/***************************************************************************/
// This will return a list of paths and the number of paths
// This solves the replacement path for a directed graph by looking for bad
// points in the best replacement and using shortest path in these cases.
// Note;Calling routine is responsible for freeing the paths
int Graph::findReplacementPaths(Path **pathsOut, int *npathsOut)
{
	int i,e;
	int u,v;	// Points of edge e
	int npath,ret;
	EdgeKey *key;
	int n;
	Edge edge;
	Path naivePath,*paths;
	double cost_v,cost_u;
	int path_u,path_v,block_u,minblock_v,block_v;
	std::list < EdgeKey * >::iterator it;

	// Init the number of paths out in case we cannot find shortest path etc.
	(*npathsOut)=0;

	// Dynamic structures
	FibonacciHeap *Q=NULL;

	// Solve the shortest path problem completely
	n=solveShortestPath(1);		
	if (n!=OK) 	return(NOT_FOUND);


	// Solve the reverse shortest path problem
	solveReverseShortestPath();

	// Compute block information from shortest path
	computeBlock();

	// Get the number of hops in shortest path
	npath=getNpath();

	// Results storage
	paths=new Path[npath];
	assert(paths!=NULL);

	// Finally, run the algorithm
	// 1. L,R k-element arrays, Q - priority queue
#ifdef PREALLOCATE
	LinkList *L=new LinkList[npath];
	LinkList *R=new LinkList[npath];
	LinkListItem *item,*next;
#else
	std::list < EdgeKey * > *L=new std::list < EdgeKey * >[npath];
	std::list < EdgeKey * > *R=new std::list < EdgeKey * >[npath];
#endif
	assert(L!=NULL);
	assert(R!=NULL);

	Q=new FibonacciHeap;
	assert(Q!=NULL);

	// Init edge information
	for (i=0;i<p0.nedges;i++) {
		edgeKey[i].n=i;
		edgeKey[i].w=0.0;
		edgeKey[i].node=NULL;
	}

	// 2. For each edge not in path
	for (e=0;e<p0.nedges;e++) {
		// If left(e) < right(e), put e into L[left(e)] and R[right(e)]
		// e=(u,v)
		// Define left(e)=block(u)
		// Define right(e)=minblock(v)
		edge=getEdge(e);
		if (edge.cost==INF) continue;

		u=edge.p1;
		v=edge.p2;
		if (labels[u].valid==FALSE || labels[v].valid==FALSE) continue;
		path_u=getPath(u);
		path_v=getPath(v);

		// If both are on path and one away continue;
		if (path_u>=0 && path_v>=0 && abs(path_u-path_v)==1) continue;

		block_u=getBlock(u);
		minblock_v=getMinBlock(v);

		// If we cannot get to point u from x block=(-1) or 
		// if we cannot get to point v from y minblock=(-1) dont add candidate edge
		if (block_u==(-1) || minblock_v==(-1)) continue;
		if (p0.directed==TRUE) {
			block_v=labels[v].block;
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
			else if (block_v>minblock_v) {	// Bad point
				if (minblock_v<npath) L[minblock_v].push_back(&(edgeKey[e]));
				if (block_v<npath) R[block_v].push_back(&(edgeKey[e]));
			}
		}
		else {
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
		}
	}

	// 3. For i=1 to k-1
	for (i=0;i<npath-1;i++) {
		// For each edge e=(u,v) in L[i]
#ifdef PREALLOCATE
		next=L[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=L[i].begin();it!=L[i].end();it++) {
			key=(*it);
#endif

			// Calculate the weight
			edge=getEdge(key->n);
			u=edge.p1;
			v=edge.p2;
			cost_u=getCost(u);
			cost_v=getCostToY(v);
			key->w=cost_u+edge.cost+cost_v;

			// Insert (w,e) into Q
			key->node=Q->insert_item(key,key->w);

			//item=item->get_next();	// Get the next item
		}

		// For each edge e in R[i] remove from Q
#ifdef PREALLOCATE
		next=R[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=R[i].begin();it!=R[i].end();it++) {
			key=(*it);
#endif

			// Delete the node from the heap
			Q->delete_node(((FNODE *)key->node));
			key->node=NULL;

		}

		key=(EdgeKey *)(Q->get_min());
		if (key!=NULL) {
			// If we are in the directed case and have a bad point in the solution
			if (p0.directed && 
				labels[edges[key->n].p2].block>labels[edges[key->n].p2].minblock) {

				// Cant use this algorithm uses shortest path
				//put_error("Using shortest path to calculate a replacement edge");
				numBadCases++;

				// Naively solve the replacement problem
				n=solveNaiveReplacement(shortestPath.list[i],paths[i]);
			}
			else {
				// Report the minimum weight in Q as the x-y distance omitting e
				// If this is a directed graph
				paths[i].cost=key->w;

				ret=computePath(key->n,paths[i]);
				// ??? Debug
				if (ret!=OK) {
					(*pathsOut)=NULL;
					(*npathsOut)=0;
					return(NO_CHANGE);
				}
			}
		}
		else {
			// Put something in the return
			paths[i].cost=0.0;
			paths[i].nList=0;
			paths[i].list=NULL;
		}

	}

	// Delete the temporary structures
	delete Q;
	delete [] L;
	delete [] R;

	// Debug 
	if (Sys.getDebug()) {
		char pathName[128],msg[1024];
		for (i=0;i<npath-1;i++) {
			// Verify the answer
			// Make the missing edge have cost=infinity and solve the shortest path problem again
			// Compare the costs within some tolerance (0.00001)

			// Naively solve the replacement of the edge that 
			// goes from from node i in the path to node i+1
			//naivePath=solveNaiveReplacement(i);
			n=solveNaiveReplacement(shortestPath.list[i],naivePath);
			if (n!=OK || naivePath.nList==0) {
				if (paths[i].cost!=INF && paths[i].nList!=0) 
					put_error("Error:Naive path cost=INF but replacement algo routine does not");
			}
			else if (fabs(paths[i].cost-naivePath.cost)>0.00001) {
				char cost1s[128],cost2s[128];

				if (paths[i].cost==DBL_MAX) strcpy(cost1s,"Inf");
				else sprintf(cost1s,"%f",paths[i].cost);
				if (naivePath.cost==DBL_MAX) strcpy(cost2s,"Inf");
				else sprintf(cost2s,"%f",naivePath.cost);

				sprintf(pathName,"Path:%d wrong cost:%s",i,cost1s);
				formatPath(pathName,naivePath,msg);
				put_error(msg);
			}
		}
	}

	// Set the return variables
	(*pathsOut)=paths;
	(*npathsOut)=npath-1;

	return(OK);
}
/***************************************************************************/
// This will return the best replacement path
// Len is the longest length of shortest path for which we will allow replacements,
int Graph::findBestReplacementPath(int len, Path &bestPath)
{
	int i,e;
	int u,v;	// Points of edge e
	int n,npath;
	EdgeKey *key;
	Edge edge;
	Path naivePath;
	double cost_v,cost_u;
	int path_u,path_v,block_u,minblock_v,block_v;
	std::list < EdgeKey * >::iterator it;
	double min=INF;
	int bestReplacementEdge=(-1);
	int naivePathFlag=FALSE;	// Flag for whether we are using a naive path for the best path

	// Solve the shortest path problem completely
	n=solveShortestPath(1);		
	if (n!=OK) return(NOT_FOUND);

	// Solve the reverse shortest path problem
	solveReverseShortestPath();

	// Compute block and minblock information from shortest path
	computeBlock();

	// Get the number of hops in the data structure
	if (len<shortestPath.nList+1) npath=len;
	else npath=shortestPath.nList+1;

	// Finally, run the algorithm
	// 1. L,R k-element arrays, Q - priority queue
#ifdef PREALLOCATE
	LinkList *L=new LinkList[npath];
	LinkList *R=new LinkList[npath];
	LinkListItem *item,*next;
#else
	std::list < EdgeKey * > *L=new std::list < EdgeKey * >[npath];
	std::list < EdgeKey * > *R=new std::list < EdgeKey * >[npath];
#endif
	assert(L!=NULL);
	assert(R!=NULL);

	FibonacciHeap *Q=new FibonacciHeap;
	assert(Q!=NULL);

	// Init edge information
	for (i=0;i<p0.nedges;i++) {
		edgeKey[i].n=i;
		edgeKey[i].w=0.0;
		edgeKey[i].node=NULL;
	}


	// 2. For each edge not in path
	for (e=0;e<p0.nedges;e++) {
		// If left(e) < right(e), put e into L[left(e)] and R[right(e)]
		// e=(u,v)
		// Define left(e)=block(u)
		// Define right(e)=minblock(v)
		edge=edges[e];
		if (edge.cost==INF) continue;

		// Get the nodes
		u=edge.p1;
		v=edge.p2;

		// If either of the nodes has been removed from the graph continue
		if (labels[u].valid==FALSE || labels[v].valid==FALSE) continue;
		path_u=labels[u].path;
		path_v=labels[v].path;

		// If both are on path and one away continue;
		if (path_u>=0 && path_v>=0 && abs(path_u-path_v)==1) continue;

		block_u=labels[u].block;
		minblock_v=labels[v].minblock;

		// If we cannot get to point u from x block=(-1) or 
		// if we cannot get to point v from y minblock=(-1) dont add candidate edge
		if (block_u==(-1) || minblock_v==(-1)) continue;
		if (p0.directed==TRUE) {
			block_v=labels[v].block;
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
			else if (block_v>minblock_v) {	// Bad point
				if (minblock_v<npath) L[minblock_v].push_back(&(edgeKey[e]));
				if (block_v<npath) R[block_v].push_back(&(edgeKey[e]));
			}
		}
		else {
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
		}
	}

	// 3. For i=1 to k-1
	for (i=0;i<npath-1;i++) {
		// For each edge e=(u,v) in L[i]
#ifdef PREALLOCATE
		next=L[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=L[i].begin();it!=L[i].end();it++) {
			key=(*it);
#endif

			// Calculate the weight
			edge=getEdge(key->n);
			u=edge.p1;
			v=edge.p2;
			cost_u=getCost(u);
			cost_v=getCostToY(v);
			key->w=cost_u+edge.cost+cost_v;

			// Insert (w,e) into Q
			key->node=Q->insert_item(key,key->w);
		}

		// For each edge e in R[i] remove from Q
#ifdef PREALLOCATE
		next=R[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=R[i].begin();it!=R[i].end();it++) {
			key=(*it);
#endif

			// Delete the node from the heap
			Q->delete_node(((FNODE *)key->node));
			key->node=NULL;
		}

		// Report the minimum weight in Q as the x-y distance omitting e
		// Keep track of the best edge
		key=(EdgeKey *)(Q->get_min());
		if (key!=NULL) {

			// If this is a directed graph
			if (p0.directed && 
				labels[edges[key->n].p2].block>labels[edges[key->n].p2].minblock) {
				// Cant use this algorithm uses shortest path
				numBadCases++;

				// Naively solve the replacement problem
				n=solveNaiveReplacement(shortestPath.list[i],naivePath);

				// If the solution is better store it 
				if (n==OK && naivePath.cost<min) {
					naivePathFlag=TRUE;
					min=naivePath.cost;
					bestPath=naivePath;
				}
			}
			else {
				if (key->w<min) {
					min=key->w;
					bestReplacementEdge=key->n;
					naivePathFlag=FALSE;
				}

				if (Sys.getDebug()) {
					char pathName[128],msg[1024];

					// Verify the answer
					// Make the missing edge have cost=infinity and solve the shortest path problem again
					// Compare the costs within some tolerance (0.00001)
					// Naively solve the replacement of the edge that 
					// goes from from node i in the path to node i+1
					n=solveNaiveReplacement(shortestPath.list[i],naivePath);
					if (n==OK && fabs(naivePath.cost-key->w)>0.01) {
						char cost1s[128],cost2s[128];

						if (key->w==INF) strcpy(cost1s,"Inf");
						else sprintf(cost1s,"%f",key->w);
						if (naivePath.cost==INF) strcpy(cost2s,"Inf");
						else sprintf(cost2s,"%f",naivePath.cost);

						sprintf(pathName,"Best replacement path wrong cost:%s",cost1s);
						formatPath(pathName,naivePath,msg);
						put_error(msg);
					}
				}
			}
		}
	}

	// Delete the temporary structures
	delete Q;
	delete [] L;
	delete [] R;

	// If the best path is because of some replacement edge, compute it
	if (bestReplacementEdge>=0 && naivePathFlag==FALSE) {
		// Fix replacement edge problem store in pathOut
		computePath(bestReplacementEdge,bestPath);
	}
	return(OK);
}
/***************************************************************************/
// This will return a list of paths and the number of paths
// This was the old way of finding the paths but does not work for a directed graph
int Graph::oldFindReplacementPaths(Path **pathsOut, int *npathsOut)
{
	int i,e;
	int u,v;	// Points of edge e
	int npath;
	EdgeKey *key;
	int n;
	Edge edge;
	Path naivePath,*paths;
	double cost_v,cost_u;
	int path_u,path_v,block_u,minblock_v,ret;
	std::list < EdgeKey * >::iterator it;

	// Init the number of paths out in case we cannot find shortest path etc.
	(*npathsOut)=0;

	// Dynamic structures
	FibonacciHeap *Q=NULL;

	// Solve the shortest path problem completely
	n=solveShortestPath(1);		
	if (n!=OK) 	return(NOT_FOUND);

	// Solve the reverse shortest path problem
	solveReverseShortestPath();

	// Compute block information from shortest path
	computeBlock();

	// Get the number of hops in the data structure
	npath=getNpath();

	// Results storage
	paths=new Path[npath];
	assert(paths!=NULL);

	// Finally, run the algorithm
	// 1. L,R k-element arrays, Q - priority queue
#ifdef PREALLOCATE
	LinkList *L=new LinkList[npath];
	LinkList *R=new LinkList[npath];
	LinkListItem *item,*next;
#else
	std::list < EdgeKey * > *L=new std::list < EdgeKey * >[npath];
	std::list < EdgeKey * > *R=new std::list < EdgeKey * >[npath];
#endif
	assert(L!=NULL);
	assert(R!=NULL);

	Q=new FibonacciHeap;
	assert(Q!=NULL);

	// Init edge information
	for (i=0;i<p0.nedges;i++) {
		edgeKey[i].n=i;
		edgeKey[i].w=0.0;
		edgeKey[i].node=NULL;
	}

	// 2. For each edge not in path
	for (e=0;e<p0.nedges;e++) {
		// If left(e) < right(e), put e into L[left(e)] and R[right(e)]
		// e=(u,v)
		// Define left(e)=block(u)
		// Define right(e)=minblock(v)
		edge=getEdge(e);
		if (edge.cost==INF) continue;

		u=edge.p1;
		v=edge.p2;
		if (labels[u].valid==FALSE || labels[v].valid==FALSE) continue;
		path_u=getPath(u);
		path_v=getPath(v);

		// If both are on path and one away continue;
		if (path_u>=0 && path_v>=0 && abs(path_u-path_v)==1) continue;

		block_u=getBlock(u);
		minblock_v=getMinBlock(v);

		// If we cannot get to point u from x block=(-1) or 
		// if we cannot get to point v from y minblock=(-1) dont add candidate edge
		if (block_u==(-1) || minblock_v==(-1)) continue;
		if (block_u<minblock_v) {
			L[block_u].push_back(&(edgeKey[e]));
			R[minblock_v].push_back(&(edgeKey[e]));
		}
	}

	// 3. For i=1 to k-1
	for (i=0;i<npath-1;i++) {
		// For each edge e=(u,v) in L[i]
#ifdef PREALLOCATE
		next=L[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=L[i].begin();it!=L[i].end();it++) {
			key=(*it);
#endif

			// Calculate the weight
			edge=getEdge(key->n);
			u=edge.p1;
			v=edge.p2;
			cost_u=getCost(u);
			cost_v=getCostToY(v);
			key->w=cost_u+edge.cost+cost_v;

			// Insert (w,e) into Q
			key->node=Q->insert_item(key,key->w);

			//item=item->get_next();	// Get the next item
		}

		// For each edge e in R[i] remove from Q
#ifdef PREALLOCATE
		next=R[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=R[i].begin();it!=R[i].end();it++) {
			key=(*it);
#endif

			// Delete the node from the heap
			Q->delete_node(((FNODE *)key->node));
			key->node=NULL;

		}

		// Report the minimum weight in Q as the x-y distance omitting e
		key=(EdgeKey *)(Q->get_min());
		if (key!=NULL) {
			paths[i].cost=key->w;
			//edge=getEdge(key->n);

			ret=computePath(key->n,paths[i]);

			// ??? Debug
			if (ret!=OK) {
				(*pathsOut)=NULL;
				(*npathsOut)=0;
				return(NO_CHANGE);
			}
		}
		else {
			paths[i].cost=INF;
			paths[i].nList=0;
			paths[i].list=NULL;
		}
	}

	// Delete the temporary structures
	delete Q;
	delete [] L;
	delete [] R;

	// Debug 
	if (Sys.getDebug()) {
		char pathName[128],msg[1024];
		for (i=0;i<npath-1;i++) {
		// Verify the answer
		// Make the missing edge have cost=infinity and solve the shortest path problem again
		// Compare the costs within some tolerance (0.00001)
			// Naively solve the replacement of the edge that 
			// goes from from node i in the path to node i+1
			n=solveNaiveReplacement(shortestPath.list[i],naivePath);
			if (n!=OK || naivePath.nList==0) {
				if (paths[i].cost!=INF && paths[i].nList!=0) 
					put_error("Error:Naive path cost=INF but replacement algo routine does not");
			}
			else if (fabs(paths[i].cost-naivePath.cost)>0.00001) {
				char cost1s[128],cost2s[128];

				if (paths[i].cost==DBL_MAX) strcpy(cost1s,"Inf");
				else sprintf(cost1s,"%f",paths[i].cost);
				if (naivePath.cost==DBL_MAX) strcpy(cost2s,"Inf");
				else sprintf(cost2s,"%f",naivePath.cost);

				sprintf(pathName,"Path:%d wrong cost:%s",i,cost1s);
				formatPath(pathName,naivePath,msg);
				put_error(msg);
			}
		}
	}

	// Set the return variables
	(*pathsOut)=paths;
	(*npathsOut)=npath-1;

	return(OK);
}
/***************************************************************************/
// This routine will find the best alternate paths for 
// 1) the lead edge and
// 2) using the lead edge some other edge
// It will store this information in the specified heap
int Graph::findBestUnionClassPaths(BRNODE *u, BRNODE *v, 
	FibonacciHeap *h, const double prefixCost)
{
	int i,e;
	int uNode,vNode;	// Points of edge e
	int n,npath;
	EdgeKey *key;
	Edge edge;
	Path bestPath,naivePath;
	Path replacementPath;
	double cost_v,cost_u;
	int path_u,path_v,block_u,minblock_v,block_v;
	std::list < EdgeKey * >::iterator it;
	double min=INF;
	int bestReplacementEdge=(-1);
	int naivePathFlag=FALSE;	// Flag for whether we are using a naive path for the best path
	int len;

	len=v->prefixPath->nList;

	// Solve the shortest path problem completely
	n=solveShortestPath(1);		
	if (n!=OK) return(NULL);

	// Solve the full reverse shortest path problem
	solveReverseShortestPath();
	//solveReverseDijkstra(1);

	// Compute block information from shortest path
	computeBlock();

	// Get the number of hops in the data structure
	if (len+1<shortestPath.nList+1) npath=len+1;
	else npath=shortestPath.nList+1;

	// Finally, run the algorithm
	// 1. L,R k-element arrays, Q - priority queue
#ifdef PREALLOCATE
	LinkList *L=new LinkList[npath];
	LinkList *R=new LinkList[npath];
	LinkListItem *item,*next;
#else
	std::list < EdgeKey * > *L=new std::list < EdgeKey * >[npath];
	std::list < EdgeKey * > *R=new std::list < EdgeKey * >[npath];
#endif
	assert(L!=NULL);
	assert(R!=NULL);

	FibonacciHeap *Q=new FibonacciHeap;
	assert(Q!=NULL); // Assert we got memory

	// Init edge information
	for (i=0;i<p0.nedges;i++) {
		edgeKey[i].n=i;
		edgeKey[i].w=0.0;
		edgeKey[i].node=NULL;
	}

	// 2. For each edge not in path
	for (e=0;e<p0.nedges;e++) {
		// If left(e) < right(e), put e into L[left(e)] and R[right(e)]
		// e=(u,v)
		// Define left(e)=block(u)
		// Define right(e)=minblock(v)
		edge=edges[e];
		if (edge.cost==INF) continue;

		// Get the nodes
		uNode=edge.p1;
		vNode=edge.p2;

		// If either of the nodes has been removed from the graph continue
		if (labels[uNode].valid==FALSE || labels[vNode].valid==FALSE) continue;
		path_u=labels[uNode].path;
		path_v=labels[vNode].path;

		// If both are on path and one away continue;
		if (path_u>=0 && path_v>=0 && abs(path_u-path_v)==1) continue;

		block_u=labels[uNode].block;
		minblock_v=labels[vNode].minblock;

		// If we cannot get to point u from x block=(-1) or 
		// if we cannot get to point v from y minblock=(-1) dont add candidate edge
		if (block_u==(-1) || minblock_v==(-1)) continue;
		if (p0.directed==TRUE) {
			block_v=labels[vNode].block;
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
			else if (block_v>minblock_v) {	// Bad point
				if (minblock_v<npath) L[minblock_v].push_back(&(edgeKey[e]));
				if (block_v<npath) R[block_v].push_back(&(edgeKey[e]));
			}
		}
		else {
			if (block_u<minblock_v) {
				if (block_u<npath) L[block_u].push_back(&(edgeKey[e]));
				if (minblock_v<npath) R[minblock_v].push_back(&(edgeKey[e]));
			}	
		}
	}

	// 3. For i=1 to k-1
	for (i=0;i<npath-1;i++) {
		// For each edge e=(u,v) in L[i]
#ifdef PREALLOCATE
		next=L[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=L[i].begin();it!=L[i].end();it++) {
			key=(*it);
#endif

			// Calculate the weight
			edge=getEdge(key->n);
			uNode=edge.p1;
			vNode=edge.p2;
			cost_u=getCost(uNode);
			cost_v=getCostToY(vNode);
			key->w=cost_u+edge.cost+cost_v;

			// Insert (w,e) into Q
			key->node=Q->insert_item(key,key->w);
		}

		// For each edge e in R[i] remove from Q
#ifdef PREALLOCATE
		next=R[i].getFirst();
		while(FOREVER) {
			item=next;
			if (item==NULL) break;
			key=(EdgeKey *)(item->key);
			next=item->next;

#else
		for (it=R[i].begin();it!=R[i].end();it++) {
			key=(*it);
#endif

			// Delete the node from the heap
			Q->delete_node(((FNODE *)key->node));
			key->node=NULL;
		}

		// Report the minimum weight in Q as the x-y distance omitting e
		// Keep track of the best edge
		key=(EdgeKey *)(Q->get_min());
		if (key!=NULL) {

			// If this is a directed graph
			if (p0.directed && 
				labels[edges[key->n].p2].block>labels[edges[key->n].p2].minblock) {
				// Cant use this algorithm uses shortest path
				//put_error("Using shortest path to calculate a replacement edge");
				numBadCases++;

				// Naively solve the replacement problem
				n=solveNaiveReplacement(shortestPath.list[i],naivePath);

				// If the solution is better store it 
				if (n==OK && naivePath.cost<min) {
					// Store best path
					naivePathFlag=TRUE;
					min=naivePath.cost;
					bestPath=naivePath;
				}
			}
			else {
				if (key->w<min) {
					// Store best path as an edge
					min=key->w;
					bestReplacementEdge=key->n;
					naivePathFlag=FALSE;
				}

				if (Sys.getDebug()) {
					char pathName[128],msg[1024];

					// Verify the answer
					// Make the missing edge have cost=infinity and solve the shortest path problem again
					// Compare the costs within some tolerance (0.00001)
					// Naively solve the replacement of the edge that 
					// goes from from node i in the path to node i+1
					n=solveNaiveReplacement(shortestPath.list[i],naivePath);
					if (n==OK && fabs(naivePath.cost-key->w)>0.01) {
						char cost1s[128],cost2s[128];

						if (key->w==INF) strcpy(cost1s,"Inf");
						else sprintf(cost1s,"%f",key->w);
						if (naivePath.cost==INF) strcpy(cost2s,"Inf");
						else sprintf(cost2s,"%f",naivePath.cost);

						sprintf(pathName,"Best replacement path wrong cost:%s",cost1s);
						formatPath(pathName,naivePath,msg);
						put_error(msg);
					}
				}
			}
		}

		// Store shortest path for C(u)
		if (i==0) {
			// If all we have is an edge - compute the path
			if (bestReplacementEdge>=0 && naivePathFlag==FALSE) {
				// Fix replacement edge problem store in pathOut
				computePath(bestReplacementEdge,bestPath);
			}

			// If we have a best path
			if (bestPath.nList>0) {
				bestPath.cost+=prefixCost;	// Tack on the cost of the prefixpath
				addPathToHeap(h,bestPath,u,NULL);
			}

			// This path will not work for path class (u,v) 
			// so start afresh
			bestReplacementEdge=(-1);
			naivePathFlag=FALSE;
			min=INF;
		}
	}

	// Delete the temporary structures
	delete Q;	 
	delete [] L;
	delete [] R;

	// If we have an edge for the best path - compute the path
	if (bestReplacementEdge>=0 && naivePathFlag==FALSE) {
		// Fix replacement edge problem store in pathOut
		computePath(bestReplacementEdge,bestPath);
	}

	// If we have a best path that uses the lead_edge
	if (bestPath.nList>0 && bestPath.list[0]==v->prefixPath->list[0]) {
		// It must meet our requirements since it was only added if it had
		// the first edge in common with shortest path (lead edge of v)
		bestPath.cost+=prefixCost;	// Tack on the cost of the prefixpath
		addPathToHeap(h,bestPath,u,v);
	}

	else {
		if (v->prefixPath->nList>1) {
			int e=v->prefixPath->list[0];

			// We must compute a replacement path
			// Set start node - find b
			setStartNode(edges[e].p2);

			// Delete node u
			setNodeInvalid(u->node);

			// Find best path the old way
			findBestReplacementPath(len,replacementPath);

			if (replacementPath.nList>0) {
				// Add deleted edge back in
				replacementPath.pushFront(e,edges[e].cost,bestPath);
				// Add the cost of getting to the starting point
				bestPath.cost+=prefixCost;
				addPathToHeap(h,bestPath,u,v);
			}
		}
	}
	return(OK);
}
/***************************************************************************/
